import pytest
from math import fmod
from smolarith.div import Sign


@pytest.fixture
def reference_tb(sim_mod, n, d, q, r, sign):
    _, m = sim_mod

    def testbench(delay):
        yield m.inp.payload.n.eq(n)
        yield m.inp.payload.d.eq(d)
        yield m.inp.payload.sign.eq(sign)
        yield m.inp.valid.eq(1)
        yield

        yield m.inp.valid.eq(0)  # Only schedule one xfer.
        yield m.outp.ready.eq(1)  # Immediately ready for retrieval.
        yield
        for _ in range(delay):
            yield

        assert (yield m.outp.valid) == 1
        assert (yield m.outp.payload.sign) == sign.value

        if (yield m.outp.payload.sign) == Sign.UNSIGNED.value:
            assert (yield m.outp.payload.q) == q
            assert (yield m.outp.payload.r) == r
        else:
            assert (yield m.outp.payload.q.as_signed()) == q
            assert (yield m.outp.payload.r.as_signed()) == r

        yield
        assert (yield m.outp.valid) == 0

    return testbench


@pytest.fixture
def mismatch_tb(sim_mod, n, d, q, r, sign):
    _, m = sim_mod

    def testbench(delay):
        yield m.inp.payload.n.eq(n)
        yield m.inp.payload.d.eq(d)
        yield m.inp.payload.sign.eq(sign)
        yield m.inp.valid.eq(1)
        yield

        yield m.inp.valid.eq(0)  # Only schedule one xfer.
        yield m.outp.ready.eq(1)  # Immediately ready for retrieval.
        yield
        for _ in range(delay):
            yield

        assert (yield m.outp.valid) == 1
        assert (yield m.outp.payload.sign) == sign.value

        if sign == Sign.UNSIGNED:
            assert (yield m.outp.payload.q) == q
            assert (yield m.outp.payload.r) == r
        else:
            assert (yield m.outp.payload.q.as_signed()) == q
            assert (yield m.outp.payload.r.as_signed()) == r

        yield
        assert (yield m.outp.valid) == 0

    return testbench


@pytest.fixture
def riscv_tb(sim_mod, n, d, q, r):
    _, m = sim_mod

    def testbench(delay):
        yield m.inp.payload.n.eq(n)
        yield m.inp.payload.d.eq(d)
        yield m.inp.payload.sign.eq(Sign.SIGNED)
        yield m.inp.valid.eq(1)
        yield

        yield m.inp.valid.eq(0)  # Only schedule one xfer.
        yield m.outp.ready.eq(1)  # Immediately ready for retrieval.
        yield
        for _ in range(delay):
            yield

        assert (yield m.outp.valid) == 1
        assert (yield m.outp.payload.q.as_signed()) == q
        assert (yield m.outp.payload.r.as_signed()) == r
        assert (yield m.outp.payload.sign) == Sign.SIGNED.value

        yield
        assert (yield m.outp.valid) == 0

    return testbench


@pytest.fixture
def signed_tb(sim_mod):
    _, m = sim_mod

    def testbench(delay):
        for n in range(-2**(m.width-1), 2**(m.width-1)):
            yield m.inp.payload.n.eq(n)
            for d in range(-2**(m.width-1), 2**(m.width-1)):
                yield m.inp.payload.d.eq(d)
                yield m.inp.payload.sign.eq(Sign.SIGNED)
                yield m.inp.valid.eq(1)
                yield

                yield m.inp.valid.eq(0)  # Only schedule one xfer.
                yield m.outp.ready.eq(1)  # Immediately ready for retrieval.
                yield
                for _ in range(delay):
                    yield

                assert (yield m.outp.valid) == 1
                assert (yield m.outp.payload.sign) == Sign.SIGNED.value
                if n == -2**(m.width-1) and d == -1:
                    assert (yield m.outp.payload.q.as_signed()) == \
                        -2**(m.width-1)
                    assert (yield m.outp.payload.r.as_signed()) == 0
                elif d == 0:
                    assert (yield m.outp.payload.q.as_signed()) == -1
                    assert (yield m.outp.payload.r.as_signed()) == n
                else:
                    assert (yield m.outp.payload.q.as_signed()) == int(n / d)
                    assert (yield m.outp.payload.r.as_signed()) == fmod(n, d)

    return testbench


@pytest.fixture
def unsigned_tb(sim_mod):
    _, m = sim_mod

    def testbench(delay):
        for n in range(0, 2**m.width):
            yield m.inp.payload.n.eq(n)
            for d in range(0, 2**m.width):
                yield m.inp.payload.d.eq(d)
                yield m.inp.payload.sign.eq(Sign.UNSIGNED)
                yield m.inp.valid.eq(1)
                yield

                yield m.inp.valid.eq(0)  # Only schedule one xfer.
                yield m.outp.ready.eq(1)  # Immediately ready for retrieval.
                yield
                for _ in range(delay):
                    yield

                assert (yield m.outp.valid) == 1
                assert (yield m.outp.payload.sign) == Sign.UNSIGNED.value
                if d == 0:
                    assert (yield m.outp.payload.q.as_unsigned()) == \
                        2**m.width - 1
                    assert (yield m.outp.payload.r.as_unsigned()) == n
                else:
                    assert (yield m.outp.payload.q.as_unsigned()) == int(n / d)
                    assert (yield m.outp.payload.r.as_unsigned()) == fmod(n, d)

    return testbench






import pytest
from functools import partial
from smolarith.div import Sign, MulticycleDiv, Impl


@pytest.mark.module(MulticycleDiv(12, impl=Impl.NON_RESTORING))
@pytest.mark.clks((1.0 / 12e6,))
@pytest.mark.parametrize("n,d,q,r,sign", [(1362, 14, 97, 4, Sign.SIGNED),
                                          (-1362, 14, -97, -4, Sign.SIGNED),
                                          (1362, -14, -97, 4, Sign.SIGNED),
                                          (-1362, -14, 97, -4, Sign.SIGNED),
                                          (1362, 14, 97, 4, Sign.UNSIGNED),])
def test_reference_div(sim_mod, reference_tb):
    sim, _ = sim_mod
    sim.run(sync_processes=[partial(reference_tb, 16 - 1)])


@pytest.mark.module(MulticycleDiv(12, impl=Impl.NON_RESTORING))
@pytest.mark.clks((1.0 / 12e6,))
@pytest.mark.parametrize("n,d,q,r,sign", [
    # 100000000001
    (2049, 2, 1024, 1, Sign.UNSIGNED),
    # 100000000001
    (-2047, 2, -1023, -1, Sign.SIGNED),
    # 100000000000
    (2048, 2, 1024, 0, Sign.UNSIGNED),
    # 100000000000
    (-2048, 2, -1024, 0, Sign.SIGNED),
    # 100000000000
    (2048, 1, 2048, 0, Sign.UNSIGNED),
    # 100000000000
    (-2048, 1, -2048, 0, Sign.SIGNED),
    # 011111111111
    (2047, 2, 1023, 1, Sign.UNSIGNED),
    # 011111111111
    (2047, 2, 1023, 1, Sign.SIGNED)])
def test_signed_unsigned_mismatch(sim_mod, mismatch_tb):
    sim, _ = sim_mod
    sim.run(sync_processes=[partial(mismatch_tb, 16 - 1)])


@pytest.mark.module(MulticycleDiv(32, impl=Impl.NON_RESTORING))
@pytest.mark.clks((1.0 / 12e6,))
@pytest.mark.parametrize("n,d,q,r", [(-(2**31), -1, -(2**31), 0),
                                     (1, 0, -1, 1),
                                     (-1, 0, -1, -1),
                                     (0xff, 0, -1, 0xff)])
def test_riscv_compliance(sim_mod, riscv_tb):
    sim, _ = sim_mod
    sim.run(sync_processes=[partial(riscv_tb, 36 - 1)])


@pytest.mark.module(MulticycleDiv(8, impl=Impl.NON_RESTORING))
@pytest.mark.clks((1.0 / 12e6,))
def test_div_8bit_signed(sim_mod, signed_tb):
    sim, _ = sim_mod
    sim.run(sync_processes=[partial(signed_tb, 12 - 1)])


@pytest.mark.module(MulticycleDiv(8, impl=Impl.NON_RESTORING))
@pytest.mark.clks((1.0 / 12e6,))
def test_div_8bit_unsigned(sim_mod, unsigned_tb):
    sim, m = sim_mod
    sim.run(sync_processes=[partial(unsigned_tb, 12 - 1)])



import pytest

from amaranth.sim import Simulator


def pytest_addoption(parser):
    parser.addoption(
        "--vcds",
        action="store_true",
        help="generate Value Change Dump (vcds) from simulations",
    )


class SimulatorFixture:
    def __init__(self, req, cfg):
        self.mod = req.node.get_closest_marker("module").args[0]
        self.name = req.node.name
        self.sim = Simulator(self.mod)
        self.vcds = cfg.getoption("vcds")

        for clk in req.node.get_closest_marker("clks").args[0]:
            self.sim.add_clock(clk)

    def run(self, sync_processes, processes=[]):
        for s in sync_processes:
            self.sim.add_sync_process(s)

        for p in processes:
            self.sim.add_process(p)

        if self.vcds:
            with self.sim.write_vcd(self.name + ".vcd", self.name + ".gtkw"):
                self.sim.run()
        else:
            self.sim.run()


@pytest.fixture
def sim_mod(request, pytestconfig):
    simfix = SimulatorFixture(request, pytestconfig)
    return (simfix, simfix.mod)



import pytest
from smolarith.mul import MulticycleMul, Sign
from itertools import product


def mk_unpipelined_testbench(m, abs_iter):
    def testbench():
        # Not yielding to the simulator when values don't change
        # can result in significant speedups.
        a_prev = None
        b_prev = None
        s_prev = None

        yield m.inp.valid.eq(1)
        yield m.outp.ready.eq(1)

        for (a, b, s) in abs_iter:
            if s == Sign.UNSIGNED:
                if a != a_prev:
                    yield m.inp.payload.a.eq(a)
                if b != b_prev:
                    yield m.inp.payload.b.eq(b)
            else:
                if a != a_prev:
                    yield m.inp.payload.a.as_signed().eq(a)
                if b != b_prev:
                    yield m.inp.payload.b.as_signed().eq(b)

            if s != s_prev:
                yield m.inp.payload.sign.eq(s)
            yield m.inp.valid.eq(1)
            yield
            (a_prev, b_prev, s_prev) = (a, b, s)

            yield m.inp.valid.eq(0)
            while not (yield m.outp.valid):
                yield

            if (yield m.outp.payload.sign) == Sign.UNSIGNED.value:
                assert a*b == (yield m.outp.payload.o)
            else:
                assert a*b == (yield m.outp.payload.o.as_signed())

    return testbench


@pytest.fixture
def all_values_tb(request, sim_mod):
    _, m = sim_mod
    mode = request.param

    if mode == Sign.UNSIGNED:
        a_range = range(0, 2**m.width)
        b_range = range(0, 2**m.width)
    elif mode == Sign.SIGNED:
        a_range = range(-2**(m.width-1), 2**(m.width-1))
        b_range = range(-2**(m.width-1), 2**(m.width-1))
    else:  # Sign.SIGNED_UNSIGNED
        a_range = range(-2**(m.width-1), 2**(m.width-1))
        b_range = range(0, 2**m.width)

    return mk_unpipelined_testbench(m, product(a_range, b_range, (mode,)))


@pytest.mark.module(MulticycleMul(6))
@pytest.mark.clks((1.0 / 12e6,))
@pytest.mark.parametrize("all_values_tb", [Sign.UNSIGNED, Sign.SIGNED,
                                           Sign.SIGNED_UNSIGNED],
                         indirect=True)
def test_unpipelined_mul(sim_mod, all_values_tb):
    sim, m = sim_mod
    sim.run(sync_processes=[all_values_tb])



"""Soft-core multiplier components."""

from amaranth import Module, Signal, signed, unsigned
from amaranth.lib.data import ArrayLayout, StructLayout
from amaranth.lib.wiring import Signature, In, Out, Component

from amaranth.lib.enum import IntEnum, auto


class Sign(IntEnum):
    """Indicate the type of multiply to be performed.

    * ``UNSIGNED``: Both inputs ``a`` and ``b`` are unsigned.
      
      The output is unsigned.

    * ``SIGNED``: Both inputs ``a`` and ``b`` are unsigned.
      
      The output is signed.

    * ``SIGNED_UNSIGNED``: Input ``a`` is signed and input ``b`` is unsigned.
      
      The output is signed.
       
      Note that for :math:`n`-bit multiply with given bit patterns for ``a``
      and ``b``, the bottom :math:`n/2` bits will be identical in an
      ``UNSIGNED`` or ``SIGNED_UNSIGNED`` multiply.
    """

    UNSIGNED = auto()
    SIGNED = auto()
    SIGNED_UNSIGNED = auto()


class Inputs(StructLayout):
    r"""Tagged union representing the signedness of multiplier inputs.
    
    * When :attr:`sign` is ``UNSIGNED``, both :attr:`a` and :attr:`b` are
      treated by the multiplier as :class:`~amaranth:amaranth.hdl.Value`\ s
      with an :data:`~amaranth:amaranth.hdl.unsigned`
      :class:`~amaranth:amaranth.hdl.Shape`.

    * When ``sign`` is ``SIGNED``, both ``a`` and ``b`` are treated as Values
      with a :data:`~amaranth:amaranth.hdl.signed` Shape.

    * When ``sign`` is ``SIGNED_UNSIGNED``, ``a`` is treated as a Value with a
      signed Shape, while ``b`` is treated as a Value with an unsigned Shape.

    Parameters
    ----------
    width : int
        Width in bits of both inputs ``a`` and ``b``. For signed
        multiplies, this includes the sign bit.

    Attributes
    ----------
    sign: Sign
        Controls the interpretation of the bit patterns of ``a`` and ``b``
        during multiplication.
    a: Signal(width)
        The multiplicand; i.e. the ':math:`a`' in :math:`a * b`.
    b: Signal(width)
        The multiplier; i.e. the ':math:`b`' in :math:`a * b`.
    """

    def __init__(self, width):
        super().__init__({
            "sign": Sign,
            "a": unsigned(width),
            "b": unsigned(width),
        })


class Outputs(StructLayout):
    """Tagged union representing the signedness of multiplier output.
    
    * When :attr:`sign` is ``UNSIGNED``, :attr:`o` should be treated as a
      :class:`~amaranth:amaranth.hdl.Value` with an
      :class:`~amaranth:amaranth.hdl.as_unsigned`
      :class:`~amaranth:amaranth.hdl.Shape`.

    * When :attr:`sign` is ``SIGNED`` or ``SIGNED_UNSIGNED``, ``o`` should
      be treated as a Value with a :class:`~amaranth:amaranth.hdl.as_signed`
      Shape.

    Parameters
    ----------
    width : int
        Width in bits of the output ``o``. For signed multiplies, this
        includes the sign bit.

    Attributes
    ----------
    sign: Sign
        Indicates whether the multiply that produced this product was signed,
        unsigned, or signed-unsigned.
    o: Signal(width)
        The product of :math:`a * b`.
    """

    def __init__(self, width):
        super().__init__({
            "sign": Sign,
            "o": unsigned(width),
        })


def multiplier_input_signature(width):
    """Create a parametric multiplier input port.

    This function returns a :class:`~amaranth:amaranth.lib.wiring.Signature`
    that's usable as a transfer initiator to a multiplier. A multiply starts
    on the current cycle when both ``valid`` and ``rdy`` are asserted.

    Parameters
    ----------
    width : int
        Width in bits of the inputs ``a`` and ``b``. For signed multiplies,
        this includes the sign bit.

    Returns
    -------
    :class:`~amaranth:amaranth.lib.wiring.Signature`
        :class:`~amaranth:amaranth.lib.wiring.Signature` containing the
        following members:

        .. attribute:: .payload
            :type: Out(Inputs)
            :noindex:

            Data input to multiplier.

        .. attribute:: .ready
            :type: In(1)
            :noindex:

            When ``1``, indicates that multiplier is ready.

        .. attribute:: .valid
            :type: Out(1)
            :noindex:

            When ``1``, indicates that multiplier data input is valid.
    """
    return Signature({
        "payload": Out(Inputs(width)),
        "ready": In(1),
        "valid": Out(1)
    })


def multiplier_output_signature(width):
    """Create a parametric multiplier output port.

    This function returns a :class:`~amaranth:amaranth.lib.wiring.Signature`
    that's usable as a transfer initiator **from** a multiplier.

    .. note:: For a core responding **to** a multiplier, which is the typical
              use case, you will want to use this Signature with the
              :data:`~amaranth:amaranth.lib.wiring.In` flow, like so:

              .. doctest::

                  >>> from smolarith.mul import multiplier_output_signature
                  >>> from amaranth.lib.wiring import Signature, In
                  >>> my_receiver_sig = Signature({
                  ...     "inp": In(multiplier_output_signature(width=8))
                  ... })

    Parameters
    ----------
    width : int
        Width in bits of output ``o``. For signed multiplies, this includes the
        sign bit.

    Returns
    -------
    :class:`~amaranth:amaranth.lib.wiring.Signature`
        :class:`~amaranth:amaranth.lib.wiring.Signature` containing the
        following members:

        .. attribute:: .payload
            :type: Out(Outputs)
            :noindex:

            Data output **from** multiplier.

        .. attribute:: .ready
            :type: In(1)
            :noindex:

            When ``1``, indicates that responder is ready to receive results
            from multiplier.

        .. attribute:: .valid
            :type: Out(1)
            :noindex:

            When ``1``, indicates that multiplier output data input is valid.
    """
    return Signature({
        "payload": Out(Outputs(width)),
        "ready": In(1),
        "valid": Out(1)
    })



class PipelinedMul(Component):
    r"""Multiplier soft-core which pipelines inputs.
     
    This multiplier core has pipeline registers that stores intermediate
    results for up to ``width`` multiplies at once. Basic control flow is
    implemented:
     
    * A multiply starts on the current cycle when both ``inp.valid`` and
      ``inp.ready`` are asserted.
    * A multiply result is available when ``outp.valid`` is asserted. The
      result is read/overwritten once a downstream core asserts ``outp.ready``.
    * ``inp.ready`` de-asserts on any cycle where ``outp.valid`` is asserted
      but ``outp.ready`` is *not* asserted. This is a pipeline *stall*.
    
    * Latency: Multiply Results for a given multiply will be available
      ``width`` clock cycles after the multiplier has seen those inputs,
      assuming no stalls.
       
      If stalls occur (``outp.valid`` is asserted while ``outp.ready`` is
      unasserted), latency increases by the length of the stalls in clock
      cycles while the given multiply was in the pipeline.
    
    * Throughput: One multiply maximum is finished per clock cycle.

    Parameters
    ----------
    width : int
        Width in bits of both inputs ``a`` and ``b``. For signed
        multiplies, this includes the sign bit. Output ``o`` width will
        be :math:`2*n`.
    debug : int, optional
        Enable debugging signals.

    Attributes
    ----------
    width : int
        Bit width of the inputs ``a`` and ``b``. Output ``o`` width will
        be :math:`2*n`.
    inp : In(multiplier_input_signature(width))
        Input interface to the multiplier.
    outp : Out(multiplier_output_signature(width))
        Output interface of the multiplier.
    debug: bool
        Flag which indicates whether internal debugging :class:`Signal`\s are
        enabled or not.

    Notes
    -----
    * This multiplier is a naive shift-add implementation, similar to how
      pen-and-pad multiplication in base-2/10 works. Internally, the multiplier
      treats the multiplier ``a`` as signed and the multiplicand ``b`` as
      unsigned.
     
      ``a``'s signedness only matters for the most-negative value possible for
      a given bit-width :math:`n`, where twos-complementing would not change
      the bit-pattern. Therefore, ``a`` is
      :ref:`automatically <amaranth:lang-widthext>` sign-extended to
      :math:`n + 1` bits in the input stage before any further processing.
     
      If ``b`` is negative, both ``b`` and ``a`` are twos-complemented in the
      input stage; since :math:`a * b = -a * -b`, no inverse transformation on
      the output stage is needed.

    * For an :math:`n`-bit multiply, this multiplier requires :math:`O(n^2)`
      storage elements (to store intermediate results).

    * The pipeline will happily perform multiplies on inputs where
      ``inp.valid`` is not asserted; the core will *not* assert ``outp.valid``
      when such multiplies reach the output interface (and thus they will be
      discarded).

    * For simplicity of implementation, and under the assumption that
      stalls will be rare, a pipeline stall stops the entire core. The core
      does not attempt to fill pipeline stages if the output interface isn't
      ready.

    Future Directions
    -----------------

    * It is :ref:`possible <karatsuba>` to implement a :math:`2*n`-bit
      multiply using 3 :math:`n`-bit multiplies. Since this library is about
      *smol* arithmetic, it may be worthwhile to create classes for Karatsuba
      multipliers.

      Larger multpliers using a smaller :class:`PipelinedMul` will still
      potentially be quite fast :).

    * Stalls stop the entire pipeline because validity information is not
      forwarded to earlier pipeline stages. Latency in the presence of stalls
      could be reduced by quashing invalid multiplies at other points in the
      pipeline besides the output. It is not much code complexity to
      add this, but impact on timing and size is not clear.
    """  

    def __init__(self, width=16, debug=False):
        self.width = width
        super().__init__({
            "inp": In(multiplier_input_signature(self.width)),
            "outp": Out(multiplier_output_signature(2*self.width))
        })
        self.debug = debug

    #   Unsigned, signed-unsigned case. The actual formula is the same
    #   in both cases (aside from interpretation of the bit patterns), because
    #   adding shifted copies of "x" works the same regardless of whether "x"
    #   is positive or negative:
    #
    #   twoscomp(x) + twoscomp(y) = (2^n - x) + (2^n - y) =
    #   2^(n + 1) - (x + y) = twocomp(x + y) with an additional bit.
    #
    #                   x7x6x5x4x3x2x1x0
    #                 * y7y6y5y4y3y2y1y0
    #                   ----------------
    #                   x7x6x5x4x3x2x1x0: z0 = x7x6x5x4x3x2x1x0*y0
    #                 x7x6x5x4x3x2x1x0: z1 = (x7x6x5x4x3x2x1x0*y1 << 1) + z0
    #               x7x6x5x4x3x2x1x0: z2 = (x7x6x5x4x3x2x1x0*y2 << 2) + z1
    #             x7x6x5x4x3x2x1x0: z3 = (x7x6x5x4x3x2x1x0*y3 << 3) + z2
    #           x7x6x5x4x3x2x1x0: z4 = (x7x6x5x4x3x2x1x0*y4 << 4) + z3
    #         x7x6x5x4x3x2x1x0: z5 = (x7x6x5x4x3x2x1x0*y5 << 5) + z4
    #       x7x6x5x4x3x2x1x0: z6 = (x7x6x5x4x3x2x1x0*y6 << 6) + z5
    #   + x7x6x5x4x3x2x1x0: z7 = (x7x6x5x4x3x2x1x0*y7 << 7) + z6
    #
    #   Signed case:
    #
    #   z8z7z6z5z4z3z2z1z0 = {          x7x7x6x5x4x3x2x1x0  if y7 is 0;
    #                          twoscomp(x7x7x6x5x4x3x2x1x0) if y7 is 1;
    #                                   ^--Not a typo! Sign-extend x7 to "x8"!
    #                                   z8 == z7, except when x == 0x80
    #                        }
    #   w7w6w5w4w3w2w1w0 = {          y7y6y5y4y3y2y1y0  if y7 is 0;
    #                        twoscomp(y7y6y5y4y3y2y1y0) if y7 is 1;
    #                      }
    #
    #   x, y, and z are signed, w is unsigned (or rather, "we don't care").
    #
    #                     x7x6x5x4x3x2x1x0
    #                   * y7y6y5y4y3y2y1y0
    #                     ----------------
    #                   z8z7z6z5z4z3z2z1z0: v0 = z8z7z6z5z4z3z2z1z0*w0
    #                 z8z7z6z5z4z3z2z1z0: v1 = (z8z7z6z5z4z3z2z1z0*w1 << 1) + v0  # noqa: E501
    #               z8z7z6z5z4z3z2z1z0: v2 = (z8z7z6z5z4z3z2z1z0*w2 << 2) + v1
    #             z8z7z6z5z4z3z2z1z0: v3 = (z8z7z6z5z4z3z2z1z0*w3 << 3) + v2
    #           z8z7z6z5z4z3z2z1z0: v4 = (z8z7z6z5z4z3z2z1z0*w4 << 4) + v3
    #         z8z7z6z5z4z3z2z1z0: v5 = (z8z7z6z5z4z3z2z1z0*w5 << 5) + v4
    #       z8z7z6z5z4z3z2z1z0: v6 = (z8z7z6z5z4z3z2z1z0*w6 << 6) + v5
    #   + z8z7z6z5z4z3z2z1z0: v7 = (z8z7z6z5z4z3z2z1z0*w7 << 7) + v6

    def elaborate(self, platform):  # noqa: D102
        def probe_pipeline_stage(i):
            if i == 0:
                stage_out = Signal.like(pipeline_out[0])
                m.d.comb += stage_out.eq(pipeline_out[0])
                self.pin = pipeline_in
                self.pout = pipeline_out
            else:
                stage_ina = Signal.like(self.pin[i - 1].a)
                stage_inb = Signal.like(self.pin[i - 1].b)
                stage_ins = Signal.like(self.pin[i - 1].s)
                m.d.comb += stage_ina.eq(self.pin[i - 1].a)
                m.d.comb += stage_inb.eq(self.pin[i - 1].b)
                m.d.comb += stage_ins.eq(self.pin[i - 1].s)

                stage_out = Signal.like(self.pout[i])
                m.d.comb += stage_out.eq(self.pout[i])

        m = Module()

        # self.a.width + 1 is required for when "a" is the negative-most value
        # for the width and self.b is also negative. Note that this can only
        # happen for SIGNED multiplies. In this scenario (see below), we need
        # to twos-complement "a". Consider an 8-bit signed signal with value
        # 0x80. The twos-complement is: ~0x80 + 1 = 0x7F + 1 = 0x80.
        #
        # In other words, the same value we started with. This also applies to
        # 0, there's no negative 0 in twos complement.
        #
        # To work around this, add an extra bit to your signed signal,
        # sign-extended automatically by Amaranth:
        # ~0x180 + 1 = 0x07F + 1 = 0x080
        #
        # This allows the twos complement conversion of "a" (done when "b" is
        # negative) be treated as positive for all possible values of "a".
        pipeline_in = Signal(
            ArrayLayout(
                StructLayout({
                    "a": signed(self.inp.payload.a.shape().width + 1),
                    "b": unsigned(self.inp.payload.b.shape().width),
                    "s": Sign,
                    "v": unsigned(1)
                }),
                self.width
            )
        )

        # FIXME: Does not work as intended. It is ignored.
        # for i in range(self.width):
        #     pipeline_in[i].s.reset = Sign.UNSIGNED

        # Relies on the optimizer to realize that not all 2*self.width^2
        # bits are actually used, regardless of whether we're shift-and-adding
        # positive or negative sign-extended (or positive zero-extended)
        # numbers:
        #
        # For example, consider these 2-bit signed numbers sign-extended
        # to 3-bits and added together, resulting in 3-bit signed numbers.
        # There are three possible cases a shift-add stage can see:
        #
        #               Positive:    Negative:   Negative, next add suppressed:  # noqa: E501
        # Accumulator:  000          11?         11?
        # Incoming:   + 00?        + 11?       + 000
        #             -----        -----       -----
        # Result Acc:   0??          1??         11?
        #
        # The sign bit in the result retains it's value in all cases, and
        # so (hopefully) the optimizer knows there's no reason to calculate it
        # for certain bits and instead just propogate the sign bit from input
        # stage.
        pipeline_out = Signal(ArrayLayout(signed(self.width*2), self.width))
        probe_pipeline_stage(0)

        # If the output isn't valid, we can accept another multiply. If
        # the output _is_ valid, we can only accept accept another multiply
        # if the output is being read this cycle.
        m.d.comb += self.inp.ready.eq(self.outp.valid.implies(self.outp.ready))
        m.d.sync += pipeline_in[0].v.eq(0)

        with m.If(self.inp.ready & self.inp.valid):
            m.d.sync += [
                pipeline_in[0].s.eq(self.inp.payload.sign),
                pipeline_in[0].v.eq(1)
            ]

            # If the multiplier is negative, and both inputs are signed,
            # we need to twos-complement the inputs. Since the inputs are
            # otherwise unmodified, we only need do this in the input stage.
            with m.If((self.inp.payload.sign == Sign.SIGNED) &
                      self.inp.payload.b.as_signed()[-1]):
                m.d.sync += [
                    # If multiplier is negative, then we need to _subtract_ the
                    # multiplicand from 0... which is the same as adding the
                    # twos complement! This also works if the multiplicand is
                    # negative!
                    pipeline_in[0].a.eq(-self.inp.payload.a.as_signed()),
                    # In twos complement, each bit doesn't directly correspond
                    # to whether an add should be suppressed or not; the twos
                    # complement of the value does! This also works for the
                    # negative-most multiplier, since we don't care about
                    # signed-ness when querying each individual bit.
                    pipeline_in[0].b.eq(-self.inp.payload.b.as_signed())
                ]
                m.d.sync += pipeline_out[0].eq(
                    -self.inp.payload.a.as_signed() *
                    (-self.inp.payload.b.as_signed())[0])

            # If multiplier is positive, then pass through the multiplicand
            # and multiplier unchanged- zero-extend if unsigned, sign-extend if
            # negative.
            # Aside from sign-extension (to accommodate the most-negative value
            # of the multiplicand in the above branch), the multiplicand's
            # signedness doesn't matter. From the multiplier's POV, either
            # we're adding positive numbers or adding negative numbers
            # together.
            with m.Else():
                # Quash sign-extension of "a" if unsigned multiply.
                with m.If(self.inp.payload.sign == Sign.UNSIGNED):
                    m.d.sync += [
                        pipeline_in[0].a.eq(self.inp.payload.a),
                        pipeline_in[0].b.eq(self.inp.payload.b)
                    ]
                    m.d.sync += pipeline_out[0].eq(self.inp.payload.a *
                                                   self.inp.payload.b[0])
                with m.Else():
                    m.d.sync += [
                        pipeline_in[0].a.eq(self.inp.payload.a.as_signed()),
                        pipeline_in[0].b.eq(self.inp.payload.b)
                    ]
                    m.d.sync += pipeline_out[0].eq(self.inp.payload.a.as_signed() *  # noqa: E501
                                                   self.inp.payload.b[0])

        # With the above out of the way, the main multiplying stages should
        # be identical regardless of signedness (because adding shift copies
        # works the same regardless of sign). Only the interpretation of
        # the bit patterns differ, depending on signedness.   
        for i in range(1, self.width):
            if self.debug:
                probe_pipeline_stage(i)

            with m.If(self.inp.ready):
                # This relies on the optimizer realizing we're doing a mul by a
                # 1 bit number (pipeline_in[i - 1].b[i]) with leading zeros.
                a = pipeline_in[i - 1].a
                b = pipeline_in[i - 1].b[i]
                acc = pipeline_out[i - 1]

                # Don't gate calculation on valid signal, it'll be discarded
                # anyway once we reach the end of the pipeline.
                m.d.sync += [
                    pipeline_in[i].eq(pipeline_in[i - 1]),
                    pipeline_out[i].eq(((a * b) << i) + acc)
                ]

        m.d.comb += self.outp.payload.o.eq(pipeline_out[self.width - 1])
        m.d.comb += self.outp.payload.sign.eq(pipeline_in[self.width - 1].s)
        m.d.comb += self.outp.valid.eq(pipeline_in[self.width - 1].v)

        return m


class MulticycleMul(Component):
    r"""Multicycle multiplier soft-core.

    This multiplier core is a gateware implementation of shift-add
    multiplication.

    * A multiply starts on the current cycle when both ``inp.valid`` and
      ``inp.ready`` are asserted.
    * A multiply result is available when ``outp.valid`` is asserted. The
      result is read/overwritten once a downstream core asserts ``outp.ready``.

    * Latency: Multiply Results for a will be available ``width`` clock cycles
      after assertion of both ``inp.valid`` and ``inp.ready``.

    * Throughput: One multiply maximum is finished every ``width`` clock
      cycles.

    Parameters
    ----------
    width : int
        Width in bits of both inputs ``a`` and ``b``. For signed
        multiplies, this includes the sign bit. Output ``o`` width will
        be :math:`2*n`.

    Attributes
    ----------
    width : int
        Bit width of the inputs ``a`` and ``b``. Output ``o`` width will
        be :math:`2*n`.
    inp : In(multiplier_input_signature(width))
        Input interface to the multiplier.
    outp : Out(multiplier_output_signature(width))
        Output interface of the multiplier.

    Notes
    -----
    * This multiplier is a naive shift-add implementation, similar to how
      pen-and-pad multiplication in base-2/10 works. Internally, the multiplier
      treats the multiplier ``a`` as signed and the multiplicand ``b`` as
      unsigned.
     
      ``a``'s signedness only matters for the most-negative value possible for
      a given bit-width :math:`n`, where twos-complementing would not change
      the bit-pattern. Therefore, ``a`` is
      :ref:`automatically <amaranth:lang-widthext>` sign-extended to
      :math:`n + 1` bits in the input stage before any further processing.
     
      If ``b`` is negative, both ``b`` and ``a`` are twos-complemented in the
      input stage; since :math:`a * b = -a * -b`, no inverse transformation on
      the output stage is needed.

    * For an :math:`n`-bit multiply, this multiplier requires :math:`O(3*n)`
      storage elements (to store copies of the input, output, and intermediate
      results).

    * The output product and (possibly-inverted) ``b`` input share a backing
      store. This works because only ``b``'s LSb is used each cycle and needs
      to be shifted out at the end of the cycle. Consequently, upper bits of
      ``o`` can be shifted in while lower bits of ``b`` are shifted out.
    """

    def __init__(self, width=16):
        self.width = width
        super().__init__({
            "inp": In(multiplier_input_signature(self.width)),
            "outp": Out(multiplier_output_signature(2*self.width))
        })

    def elaborate(self, platform):  # noqa: D102
        m = Module()

        # The implementation of MulticycleMul is very similar to
        # PipelinedMul, except for the lack of pipeline stages. See that
        # class for detailed notes.
        addend = Signal(signed(self.inp.payload.a.shape().width + 1))
        s_copy = Signal(Sign)

        # b and output can share a backing store; every cycle, the
        # concatenation of b and output gets shifted right after adding.
        # Output gains one bit per cycle while b loses one bit.
        #
        # The output portion of the concatenation can't tell the difference
        # between the input (to be added) being shifted left and the output
        # being shifted right before adding, as long as the backing store
        # is wide enough.
        intermediate = Signal(signed(2*self.width + 1))
        iters_left = Signal(range(self.inp.payload.b.shape().width))
        in_progress = Signal()

        m.d.comb += self.inp.ready.eq((self.outp.ready & self.outp.valid) |
                                      ~in_progress)
        m.d.comb += in_progress.eq(iters_left != 0)

        with m.If(self.outp.valid & self.outp.ready):
            m.d.sync += self.outp.valid.eq(0)

        with m.If(self.inp.ready & self.inp.valid):
            m.d.sync += [
                s_copy.eq(self.inp.payload.sign),
                iters_left.eq(self.inp.payload.b.shape().width - 1)
            ]

            with m.If((self.inp.payload.sign == Sign.SIGNED) &
                      self.inp.payload.b.as_signed()[-1]):
                # We've already done the first calculation so premptively
                # shift when initializing the backing store.
                pprod = (-self.inp.payload.a.as_signed() *
                         (-self.inp.payload.b.as_signed())[0])
                oshift = len(self.inp.payload.b)

                m.d.sync += [
                    addend.eq(-self.inp.payload.a.as_signed()),
                    # Load-bearing parens... (-self.inp.payload.b) will
                    # increase width by one to accomodate negating the most
                    # negative value. We don't actually want this because it'll
                    # disturb the backing store for the output, so suppress
                    # by using (-self.inp.payload.b)[:-1].
                    intermediate.eq(((pprod << oshift) |
                                     (-self.inp.payload.b)[:-1]) >> 1),
                ]
            with m.Else():
                # Quash sign-extension of "a" if unsigned multiply.
                with m.If(self.inp.payload.sign == Sign.UNSIGNED):
                    pprod = (self.inp.payload.a * self.inp.payload.b[0])
                    oshift = len(self.inp.payload.b)

                    m.d.sync += [
                        addend.eq(self.inp.payload.a),
                        intermediate.eq(((pprod << oshift) |
                                         self.inp.payload.b) >> 1),
                    ]
                with m.Else():
                    pprod = (self.inp.payload.a.as_signed() *
                             self.inp.payload.b[0])
                    oshift = len(self.inp.payload.b)

                    m.d.sync += [
                        addend.eq(self.inp.payload.a.as_signed()),
                        intermediate.eq(((pprod << oshift) |
                                         self.inp.payload.b) >> 1),
                    ]

        with m.If(iters_left != 0):
            m.d.sync += iters_left.eq(iters_left - 1)

            pprod = (addend * intermediate[0])
            oshift = len(self.inp.payload.b)

            m.d.sync += \
                intermediate.eq((intermediate + (pprod << oshift)) >> 1)

            with m.If(iters_left - 1 == 0):
                m.d.sync += self.outp.valid.eq(1)

        m.d.comb += [
            self.outp.payload.o.eq(intermediate),
            self.outp.payload.sign.eq(s_copy),
        ]

        return m



"""Soft-core divider components."""

from amaranth import Elaboratable, Module, Signal, signed, C, unsigned
from amaranth.lib.data import StructLayout
from amaranth.lib.wiring import Signature, In, Out, Component
from amaranth.lib.enum import Enum, auto


class Sign(Enum):
    """Indicate the type of divide to be performed.

    * ``UNSIGNED``: Both inputs ``n`` and ``d`` are unsigned.

      The output is unsigned.

    * ``SIGNED``: Both inputs ``n`` and ``d`` are unsigned.

      The quotient and remainder are signed. The remainder takes
      the sign of the dividend.
    """

    UNSIGNED = auto()
    SIGNED = auto()


class Inputs(StructLayout):
    r"""Tagged union representing the signedness of divider inputs.

    * When :attr:`sign` is ``UNSIGNED``, both :attr:`n` and :attr:`d` are
      treated by the divider as :class:`~amaranth:amaranth.hdl.Value`\ s
      with an :data:`~amaranth:amaranth.hdl.unsigned`
      :class:`~amaranth:amaranth.hdl.Shape`.

    * When ``sign`` is ``SIGNED``, both ``n`` and ``d`` are treated as Values
      with a :data:`~amaranth:amaranth.hdl.signed` Shape.

    Parameters
    ----------
    width : int
        Width in bits of both inputs ``n`` and ``d``. For signed divides, this
        includes the sign bit.

    Attributes
    ----------
    sign: Sign
        Controls the interpretation of the bit patterns of ``n`` and ``d``
        during division.
    n: Signal(width)
        The dividend; i.e. the ':math:`n`' in :math:`n / d`.
    d: Signal(width)
        The divisor; i.e. the ':math:`d`' in :math:`n / d`.
    """

    def __init__(self, width):
        super().__init__({
            "sign": Sign,
            "n": unsigned(width),
            "d": unsigned(width),
        })


class Outputs(StructLayout):
    r"""Tagged union representing the signedness of divider outputs.

    * When :attr:`sign` is ``UNSIGNED``, :attr:`q` and :attr:`r` should be
      treated as :class:`~amaranth:amaranth.hdl.Value`\ s with an
      :class:`~amaranth:amaranth.hdl.as_unsigned`
      :class:`~amaranth:amaranth.hdl.Shape`.

    * When ``sign`` is ``SIGNED``, ``q`` and ``r`` should be treated as Values
      with a :class:`~amaranth:amaranth.hdl.as_signed` Shape.

    Parameters
    ----------
    width : int
        Width in bits of the outputs ``q`` and ``r``. For signed dividers, this
        includes the sign bit.

    Attributes
    ----------
    sign: Sign
        Indicates whether the divider that produced this quotient/remainder
        was signed or unsigned.
    q: Signal(width)
        The quotient of :math:`n / d`.
    r: Signal(width)
        The remainder of :math:`n / d`, i.e. :math:`n \bmod d`.
    """

    def __init__(self, width):
        super().__init__({
            "sign": Sign,
            "q": unsigned(width),
            "r": unsigned(width)
        })


def divider_input_signature(width):
    """Create a parametric divider input port.

    This function returns a :class:`~amaranth:amaranth.lib.wiring.Signature`
    that's usable as a transfer initiator to a divider. A divider starts
    on the current cycle when both ``valid`` and ``rdy`` are asserted.

    Parameters
    ----------
    width : int
        Width in bits of the inputs :attr:`n` and :attr:`d`. For signed
        divides, this includes the sign bit.

    Returns
    -------
    :class:`~amaranth:amaranth.lib.wiring.Signature`
        :class:`~amaranth:amaranth.lib.wiring.Signature` containing the
        following members:

        .. attribute:: .payload
            :type: Out(Inputs)
            :noindex:

            Data input to divider.

        .. attribute:: .ready
            :type: In(1)
            :noindex:

            When ``1``, indicates that divider is ready.

        .. attribute:: .valid
            :type: Out(1)
            :noindex:

            When ``1``, indicates that divider data input is valid.
    """
    return Signature({
        "payload": Out(Inputs(width)),
        "ready": In(1),
        "valid": Out(1)
    })


def divider_output_signature(width):
    """Create a parametric divider output port.

    This function returns a :class:`~amaranth:amaranth.lib.wiring.Signature`
    that's usable as a transfer initiator **from** a divider.

    .. note:: For a core responding **to** a divider, which is the typical
              use case, you will want to use this Signature with the
              :data:`~amaranth:amaranth.lib.wiring.In` flow, like so:

              .. doctest::

                  >>> from smolarith.div import divider_output_signature
                  >>> from amaranth.lib.wiring import Signature, In
                  >>> my_receiver_sig = Signature({
                  ...     "inp": In(divider_output_signature(width=8))
                  ... })

    Parameters
    ----------
    width : int
        Width in bits of the outputs ``q`` and ``r``. For signed divides, this
        includes the sign bit.

    Returns
    -------
    :class:`~amaranth:amaranth.lib.wiring.Signature`
        :class:`~amaranth:amaranth.lib.wiring.Signature` containing the
        following members:

        .. attribute:: .payload
            :type: Out(Outputs)
            :noindex:

            Data output **from** divider.

        .. attribute:: .ready
            :type: In(1)
            :noindex:

            When ``1``, indicates that responder is ready to receive results
            from divider.

        .. attribute:: .valid
            :type: Out(1)
            :noindex:

            When ``1``, indicates that divider output data input is valid.
    """
    return Signature({
        "payload": Out(Outputs(width)),
        "ready": In(1),
        "valid": Out(1)
    })


class _Quadrant(StructLayout):
    """Store sign information about divider inputs."""

    def __init__(self):  # noqa: DOC
        super().__init__({
            "n": unsigned(1),
            "d": unsigned(1),
        })


class _Negate(Component):
    def __init__(self, width=8):  # noqa: DOC
        self.width = width
        super().__init__({
            "inp": In(self.width),
            "outp": Out(self.width)
        })

    
    def elaborate(self, plat):
        m = Module()

        m.d.comb += self.outp.eq((-self.inp.as_signed())[:-1])

        return m


class Impl(Enum):
    """Indicate the division algorithm to use inside a dividing component.

    * ``RESTORING``: Use :ref:`restoring division <derive-res>` for the
      internal divider. *This is a good default*.

    * ``NON_RESTORING``: Use :ref:`non-restoring division <derive-nr>` for the
      internal divider.
    """

    RESTORING = auto()
    NON_RESTORING = auto()


class MulticycleDiv(Component):
    # FIXME: I can't be .. _latency label to work, even with :ref:`latency`...
    # always "undefined label"... huh?!
    r"""Restoring/Non-restoring divider soft-core.

    This divider core is a gateware implementation of 
    :ref:`restoring division <derive-res>` or
    :ref:`non-restoring division <derive-nr>`. It works for both signed and
    unsigned divides, and should be preferred to :class:`LongDivider` in almost
    all circumstances due to better resource usage.

    * A divide starts on the current cycle when both ``inp.valid`` and
      ``inp.ready`` are asserted.
    * A divide result is available when ``outp.valid`` is asserted. The
      result is read/overwritten once a downstream core asserts ``outp.ready``.

    * Latency: Divide Results for a restoring divider will be available
      ``width + 3`` clock cycles after assertion of both ``inp.valid`` and
      ``inp.ready``. For a non-restoring divider, results are available after
      ``width + 4`` clock cycles.

    * Throughput: One divide maximum is finished every ``width + 3``
      clock cycles for a restoring divider, and every ``width + 4`` for a
      non-restoring divider.

    Parameters
    ----------
    width : int
        Width in bits of both inputs ``n`` and ``d``. For signed
        multiplies, this includes the sign bit. Outputs ``q`` and ``r`` width
        will be the same width.
    impl : Impl
        Choose whether to use a restoring divider or non-restoring divider
        internally. The default of ``RESTORING`` is fine for most cases.

    Attributes
    ----------
    width : int
        Bit width of the inputs ``n`` and ``d``, and outputs ``q`` and ``r``.
    inp : In(divider_input_signature(width))
        Input interface to the divider.
    outp : Out(divider_output_signature(width))
        Output interface of the divider.

    Notes
    -----
    * This divider is implemented using either restoring or non-restoring
      division. It is basically a gateware translation of the
      :ref:`Python <resdiv-py>` :ref:`implementations <nrdiv-py>` shown in
      :ref:`impl`.

    * Unlike the multipliers, where I could eyeball approximate storage element
      usage, divider size was benchmarked using
      `yosys <https://yosyshq.net/yosys/>`_\ 's ``synth_ice40`` pass. For an
      :math:`n`-bit divide:
    
      * ``impl=Impl.RESTORING`` requires approximately :math:`O(7*n)`
        storage elements and :math:`O(12.75*n)` LUT4s.
      * ``impl=Impl.NON_RESTORING`` requires approximately :math:`O(8*n)`
        storage elements and :math:`O(14.5*n)` LUT4s.

    * Internally, the divider converts its operands from signed to unsigned
      if necessary, performs the division, and the converts back from unsigned
      to signed if necessary. I ran into some :ref:`issues <signedness>` with
      making a signed-aware non-restoring divider such that eating the
      conversion cost latency is probably justifiable for implementation
      simplicity of both a restoring and non-restoring divider.

    .. _latency:

      The combination of signed-to-unsigned conversion, using a cycle to
      latching inputs to the internal divider, and unsigned-to-signed
      conversion adds three cycles of latency beyond the expected ``width``
      number of cycles to perform a division.

      Additionally, when using a non-restoring divider, the quotient and
      remainder require a possible :ref:`final restore step <nrdiv-restore>`.
      Remainder restore is implemented as shown in the
      :ref:`Python code <nrdiv-py>`. Quotient restore is implemented by
      subtracting the ones complement of the raw quotient from the raw quotient
      itself.

    * The quotient and remainder share a backing store; new bits are shifted
      into the lower-half quotient portion as the remainder is shifted into the
      upper half. This works because each iteration of the
      restoring/non-restoring loops check the sign of the remainder, and the
      lower quotient bits won't affect the sign bit.

    * This divider is compliant with RISC-V semantics for divide-by-zero and
      overflow when ``width=32`` or ``width=64``\ [rv]_. Specifically:

      * Signed/unsigned divide-by-zero returns "all bits set" for the
        quotient and the dividend as the remainder.
      * Signed overflow (:math:`-2^{\{31,63\}}/-1`) returns
        :math:`-2^{\{31,63\}}` for the quotient and :math:`0` as the
        remainder.

    Future Directions
    -----------------

    * It may be worthwhile to make a pipelined divider?

    * Current latency is worse than the :class:`LongDivider`, which is just
      about the only advantage to ``LongDivider``. We can provide an option to
      reduce latency/resources required due to the signed/unsigned and restore
      stages at the cost of some timing closure.
    """

    def __init__(self, width=8, impl=Impl.RESTORING):
        self.width = width
        if impl == Impl.RESTORING:
            self.div_impl = _RestoringDiv(width)
        elif impl == Impl.NON_RESTORING:
            self.div_impl = _NonRestoringDiv(width)
        self.negate_nq = _Negate(width)
        self.negate_dr = _Negate(width)

        super().__init__({
            "inp": In(divider_input_signature(self.width)),
            "outp": Out(divider_output_signature(self.width))
        })

    def elaborate(self, plat):  # noqa: D102
        m = Module()

        m.submodules.div_impl = self.div_impl
        # Time-division multiplex negation units... in a multicycle div, only
        # n/d or q/r need to be negated at once.
        m.submodules.negate_nq = self.negate_nq
        m.submodules.negate_dr = self.negate_dr

        quad = Signal(_Quadrant())
        div_inputs = Signal(Inputs(self.width))

        m.d.comb += self.div_impl.inp.payload.eq(div_inputs)

        with m.FSM():
            with m.State("IDLE"):
                m.d.comb += [
                    self.inp.ready.eq(1),
                    self.negate_nq.inp.eq(self.inp.payload.n),
                    self.negate_dr.inp.eq(self.inp.payload.d)
                ]

                m.d.comb += self.inp.ready.eq(1)
                with m.If(self.inp.valid):
                    m.next = "DIVIDE_IN"
                    
                    # Re: (self.inp.payload.d != 0)
                    # For RISCV compliance when dividing by zero, we need to
                    # suppress signed-unsigned conversion. Treating the input
                    # bit pattern as-is will result in the correct behavior of
                    # "all bits set in divisor" and remainder unchanged,
                    # regardless of sign.
                    m.d.sync += [
                        quad.n.eq(self.inp.payload.n[-1] &
                                  (self.inp.payload.d != 0)),
                        quad.d.eq(self.inp.payload.d[-1]),
                        div_inputs.eq(self.inp.payload)
                    ]

                    with m.If(self.inp.payload.sign == Sign.SIGNED):
                        with m.If(self.inp.payload.n[-1] &
                                  (self.inp.payload.d != 0)):
                            m.d.sync += div_inputs.n.eq(self.negate_nq.outp)
                        with m.If(self.inp.payload.d[-1]):
                            m.d.sync += div_inputs.d.eq(self.negate_dr.outp)

            with m.State("DIVIDE_IN"):
                m.d.comb += self.div_impl.inp.valid.eq(1)

                with m.If(self.div_impl.inp.ready):
                    m.next = "DIVIDE_LOOP"

            with m.State("DIVIDE_LOOP"):
                m.d.comb += [
                    self.div_impl.outp.ready.eq(1),
                    self.negate_nq.inp.eq(self.div_impl.outp.payload.q),
                    self.negate_dr.inp.eq(self.div_impl.outp.payload.r)
                ]

                with m.If(self.div_impl.outp.valid):
                    m.next = "DIVIDE_OUT"

                    m.d.sync += [
                        self.outp.payload.eq(self.div_impl.outp.payload),
                    ]

                    with m.If((div_inputs.sign == Sign.SIGNED) &
                              quad.n & ~quad.d):
                        m.d.sync += [
                            self.outp.payload.q.eq(self.negate_nq.outp),
                            self.outp.payload.r.eq(self.negate_dr.outp)
                        ]

                    with m.If((div_inputs.sign == Sign.SIGNED) &
                              ~quad.n & quad.d):
                        m.d.sync += self.outp.payload.q.eq(self.negate_nq.outp)
                        
                    with m.If((div_inputs.sign == Sign.SIGNED) &
                              quad.n & quad.d):
                        m.d.sync += self.outp.payload.r.eq(self.negate_dr.outp)
                        
            with m.State("DIVIDE_OUT"):
                m.d.comb += self.outp.valid.eq(1)

                with m.If(self.outp.ready):
                    m.next = "IDLE"

        return m


class _RestoringDiv(Component):
    """Unsigned-only restoring divider."""

    def __init__(self, width=8):  # noqa: DOC
        self.width = width
        # sign Signals are unused- Sign.UNSIGNED is implicit.
        super().__init__({
            "inp": In(divider_input_signature(self.width)),
            "outp": Out(divider_output_signature(self.width))
        })

    def elaborate(self, platform):
        m = Module()

        # Need extra bit because we need to subtract up to 2^width, which
        # should remain negative.
        intermediate = Signal(2*self.width + 1)
        iters_left = Signal(range(self.inp.payload.n.shape().width + 1))
        in_progress = Signal()
        s_copy = Signal(Sign)
        d_copy = Signal(self.inp.payload.d.shape())

        m.d.comb += self.inp.ready.eq((self.outp.ready & self.outp.valid) |
                                      ~in_progress)
        m.d.comb += in_progress.eq(iters_left != 0)

        with m.If(self.outp.valid & self.outp.ready):
            m.d.sync += self.outp.valid.eq(0)

        with m.If(self.inp.ready & self.inp.valid):
            m.d.sync += [
                s_copy.eq(self.inp.payload.sign),
                d_copy.eq(self.inp.payload.d),
                iters_left.eq(self.inp.payload.n.shape().width),
                intermediate.eq(self.inp.payload.n)
            ]

        with m.If(in_progress):
            # State Control
            m.d.sync += iters_left.eq(iters_left - 1)

            with m.If((iters_left - 1) == 0):
                m.d.sync += self.outp.valid.eq(1)

            # Loop iter
            # S = (S << 1) - D
            with m.If(((intermediate << 1) - (d_copy << self.width)) < 0):
                # S += D
                m.d.sync += intermediate.eq((intermediate << 1))  # noqa: E501
            with m.Else():
                # q <<=1 and then q |= 1
                # (Docs do q |= 1 and then q <<= 1, requiring a >> at the end.)
                m.d.sync += intermediate.eq(((intermediate << 1) -
                                             (d_copy << self.width)) | 1)  # noqa: E501

        m.d.comb += [
            self.outp.payload.q.eq(intermediate[:self.width]),
            self.outp.payload.r.eq(intermediate[self.width:]),
            self.outp.payload.sign.eq(s_copy)
        ]

        return m


class _NonRestoringDiv(Component):
    """Unsigned-only non-restoring divider."""

    def __init__(self, width=8):  # noqa: DOC
        self.width = width
        # sign Signals are unused- Sign.UNSIGNED is implicit.
        super().__init__({
            "inp": In(divider_input_signature(self.width)),
            "outp": Out(divider_output_signature(self.width))
        })

    def elaborate(self, platform):
        m = Module()

        # Need extra bit because we need to subtract up to 2^width, which
        # should remain negative.
        intermediate = Signal(2*self.width + 1)
        iters_left = Signal(range(self.inp.payload.n.shape().width + 1))
        in_progress = Signal()
        restore_step = Signal()
        s_copy = Signal(Sign)
        d_copy = Signal(self.inp.payload.d.shape())
        r = Signal(self.width)

        m.d.comb += self.inp.ready.eq((self.outp.ready & self.outp.valid) |
                                      ~in_progress)
        m.d.comb += in_progress.eq((iters_left != 0) | restore_step)

        with m.If(self.outp.valid & self.outp.ready):
            m.d.sync += self.outp.valid.eq(0)

        with m.If(self.inp.ready & self.inp.valid):
            m.d.sync += [
                s_copy.eq(self.inp.payload.sign),
                d_copy.eq(self.inp.payload.d),
                iters_left.eq(self.inp.payload.n.shape().width),
                intermediate.eq(self.inp.payload.n)
            ]

        with m.If(in_progress & ~restore_step):
            # State Control
            m.d.sync += iters_left.eq(iters_left - 1)

            with m.If((iters_left - 1) == 0):
                m.d.sync += restore_step.eq(1)

            # Loop iter
            with m.If(intermediate[-1]):
                # S = (S << 1) + D, encoded in top half.
                # q = -1, encoded as 0, encoded in bottom half.
                m.d.sync += intermediate.eq(((intermediate << 1) +
                                            (d_copy << self.width)))  # noqa: E501
            with m.Else():
                # S = (S << 1) - D, encoded in top half.
                # q = 1, encoded as 1, encoded in bottom half.
                m.d.sync += intermediate.eq(((intermediate << 1) -
                                            (d_copy << self.width)) | 1)  # noqa: E501


        with m.If(restore_step):
            m.d.sync += [
                restore_step.eq(0),
                self.outp.valid.eq(1)
            ]

            # Extract encoded negative powers of two and then subtract as if
            # they were positive powers of two (so no twos complement
            # conversion needed).
            # Better LUT usage if "r" has its own register.
            m.d.sync += [
                intermediate[:self.width].eq(
                    intermediate[:self.width] - ~intermediate[:self.width]),
                r.eq(intermediate[self.width:])
            ]

            with m.If(intermediate[-1]):
                # S += D
                # q -= 1
                m.d.sync += [
                    intermediate[:self.width].eq(
                        intermediate[:self.width] -
                        ~intermediate[:self.width] -
                        1),
                    r.eq(intermediate[self.width:] + d_copy)
                ]

        m.d.comb += [
            self.outp.payload.q.eq(intermediate[:self.width]),
            self.outp.payload.r.eq(r),
            self.outp.payload.sign.eq(s_copy)
        ]

        return m


class LongDivider(Component):
    r"""Long-division soft-core, used as a reference.

    This divider core is a gateware implementation of classic long division.

    * A divide starts on the current cycle when both ``inp.valid`` and
      ``inp.ready`` are asserted.
    * A divide result is available when ``outp.valid`` is asserted. The
      result is read/overwritten once a downstream core asserts ``outp.ready``.

    .. warning::

        This core is not intended to be used in user designs due to poor
        resource usage. It is mainly kept as a known-to-work reference design
        for possible equivalence checking later. Use :class:`MulticycleDiv`
        instead.

    * Latency: Divide Results for a will be available ``width`` clock cycles
      after assertion of both ``inp.valid`` and ``inp.ready``.

    * Throughput: One divide maximum is finished every ``width``
      clock cycles.

    Parameters
    ----------
    width : int
        Width in bits of both inputs ``n`` and ``d``. For signed
        multiplies, this includes the sign bit. Outputs ``q`` and ``r`` width
        will be the same width.

    Attributes
    ----------
    width : int
        Bit width of the inputs ``n`` and ``d``, and outputs ``q`` and ``r``.
    inp : In(divider_input_signature(width))
        Input interface to the divider.
    outp : Out(divider_output_signature(width))
        Output interface of the divider.

    Notes
    -----
    * This divider is a naive long-division implementation, similar to how
      pen-and-pad division is done in base-2/10.

    * Internally, the divider is aware of the sign of its inputs as well as
      the sign of the divide operation.

      The divider will dispatch to one of the 4 possible combinations of signs
      during the division loop. The four combinations are:

        * Add shifted copies of **positive**/**negative** powers of two
          together to form the quotient.
        * **Add**/**subtract** shifted copies of the divisor from the dividend
          **towards zero** to form the remainder.

    * This divider is compliant with RISC-V semantics for divide-by-zero and
      overflow when ``width=32`` or ``width=64``\ [rv]_. Specifically:

        * Signed/unsigned divide-by-zero returns "all bits set" for the
          quotient and the dividend as the remainder.
        * Signed overflow (:math:`-2^{\{31,63\}}/-1`) returns
          :math:`-2^{\{31,63\}}` for the quotient and :math:`0` as the
          remainder.

    .. [rv] https://github.com/riscv/riscv-isa-manual/releases/tag/Ratified-IMAFDQC
    """

    def __init__(self, width=8):
        self.width = width
        super().__init__({
            "inp": In(divider_input_signature(self.width)),
            "outp": Out(divider_output_signature(self.width))
        })

    def elaborate(self, platform):  # noqa: D102
        m = Module()

        m.submodules.mag = mag = _MagnitudeComparator(2*self.width)

        quotient = Signal(2*self.width)
        remainder = Signal(2*self.width)
        iters_left = Signal(range(self.width))
        in_progress = Signal()
        s_copy = Signal(Sign)
        a_sign = Signal()
        n_sign = Signal()

        # Reduce latency by 1 cycle by preempting output being read.
        m.d.comb += self.inp.ready.eq((self.outp.ready & self.outp.valid) |
                                      ~in_progress)
        m.d.comb += in_progress.eq(iters_left != 0)
        m.d.comb += [
            self.outp.payload.q.eq(quotient),
            self.outp.payload.r.eq(remainder),
            self.outp.payload.sign.eq(s_copy),
        ]

        with m.If(self.outp.ready & self.outp.valid):
            m.d.sync += self.outp.valid.eq(0)

        with m.If(self.inp.ready & self.inp.valid):
            m.d.sync += [
                # We handle first cycle using shift_amt mux.
                iters_left.eq(self.width - 1),
                a_sign.eq(self.inp.payload.n[-1]),
                n_sign.eq(self.inp.payload.d[-1]),
                s_copy.eq(self.inp.payload.sign)
            ]

            # When dividing by 0, for RISCV compliance, we need the division to
            # return -1. Without redirecting quotient calculation to the
            # "both dividend/divisor positive" case, dividing a negative
            # number by 0 returns 1. Note that the sign-bit processing doesn't
            # need to be special-cased, I do it anyway in case I see some
            # patterns I can refactor out.
            with m.If((self.inp.payload.sign == Sign.SIGNED) &
                      (self.inp.payload.d == 0) &
                      self.inp.payload.n[-1]):
                m.d.sync += a_sign.eq(0)
                m.d.sync += n_sign.eq(0)

            # We are committed to do a calc at this point, so might as well
            # start it now.
            #
            # Division quick refresh:
            #
            # Dividend / Divisor = Quotient
            # Dividend % Divisor = Remainder (takes the sign of the Dividend)
            #
            #         Quotient
            # Divisor)Dividend
            #         Remainder
            #
            # We're doing binary long division in hardware.
            # If starting with a positive dividend, we want to subtract
            # positive numbers from the dividend until we get as close to 0
            # without going below.
            # If '' negative dividend '' subtract negative numbers ''  as close
            # to 0 without going above. This becomes the remainder.
            #
            # We create the numbers to subtract from the dividend by
            # multiplying the divisor by either positive or negative powers
            # of two (shifting), depending on the sign of both the dividend
            # and divisor. When we add powers all of these two together, we
            # get the quotient.
            #
            # Either way, we can only subtract a shifted divisor from the
            # dividend if:
            #
            # * divisor*2**n <= dividend if dividend is positive.
            # * -divisor*2**n >= dividend if dividend is negative.
            #
            # We can use a magnitude comparator for this.

            shift_amt = 2**(self.width - 1)
            with m.If(self.inp.payload.sign == Sign.SIGNED):
                m.d.comb += [
                    mag.divisor.eq((self.inp.payload.d.as_signed() * shift_amt).as_signed()),  # noqa: E501
                    mag.dividend.eq(self.inp.payload.n.as_signed())
                ]
            with m.Else():
                m.d.comb += [
                    mag.divisor.eq(self.inp.payload.d.as_unsigned() *
                                   shift_amt),
                    mag.dividend.eq(self.inp.payload.n.as_unsigned())
                ]

            with m.If(mag.o):
                # If dividend/divisor are positive, subtract a positive
                # shifted divisor from dividend.
                with m.If((self.inp.payload.sign == Sign.UNSIGNED) |
                          (~self.inp.payload.n[-1] & ~self.inp.payload.d[-1]) |
                          (self.inp.payload.n[-1] & self.inp.payload.d == 0)):
                    m.d.sync += quotient.eq(C(1) * shift_amt)
                    with m.If(self.inp.payload.sign == Sign.SIGNED):
                        # If high bit is set, and, a signed div,
                        # we want to sign-extend.
                        m.d.sync += remainder.eq(self.inp.payload.n.as_signed() -  # noqa: E501
                                                 (self.inp.payload.d.as_signed() * C(1) * shift_amt).as_signed())  # noqa: E501
                    with m.Else():
                        # Otherwise, zero-extend.
                        m.d.sync += remainder.eq(self.inp.payload.n.as_unsigned() -  # noqa: E501
                                                 (self.inp.payload.d.as_unsigned() * C(1) * shift_amt))  # noqa: E501
                # If dividend is negative, but divisor is positive, create a
                # negative shifted divisor and subtract from the dividend.
                with m.If((self.inp.payload.sign == Sign.SIGNED) &
                          self.inp.payload.n[-1] & ~self.inp.payload.d[-1] &
                          ~(self.inp.payload.d == 0 & self.inp.payload.n[-1])):
                    m.d.sync += quotient.eq(C(-1) * shift_amt)
                    m.d.sync += remainder.eq(self.inp.payload.n.as_signed() -
                                             (self.inp.payload.d.as_signed() * C(-1) * shift_amt).as_signed())  # noqa: E501
                # If dividend is positive, but divisor is negative, create a
                # positive shifted divisor and subtract from the dividend.
                with m.If((self.inp.payload.sign == Sign.SIGNED) &
                          ~self.inp.payload.n[-1] & self.inp.payload.d[-1]):
                    m.d.sync += quotient.eq(C(-1) * shift_amt)
                    m.d.sync += remainder.eq(self.inp.payload.n.as_signed() -
                                             (self.inp.payload.d.as_signed() * C(-1) * shift_amt).as_signed())  # noqa: E501
                # If dividend/divisor is negative, subtract a negative
                # shifted divisor and subtract from the dividend.
                with m.If((self.inp.payload.sign == Sign.SIGNED) &
                          self.inp.payload.n[-1] & self.inp.payload.d[-1]):
                    m.d.sync += quotient.eq(C(1) * shift_amt)
                    m.d.sync += remainder.eq(self.inp.payload.n.as_signed() -
                                             (self.inp.payload.d.as_signed() * C(1) * shift_amt).as_signed())  # noqa: E501
            with m.Else():
                m.d.sync += quotient.eq(0)
                with m.If(self.inp.payload.sign == Sign.SIGNED):
                    # If high bit is set, and, a signed div,
                    # we want to sign-extend.
                    m.d.sync += remainder.eq(self.inp.payload.n.as_signed())
                with m.Else():
                    # Otherwise, zero-extend.
                    m.d.sync += remainder.eq(self.inp.payload.n.as_unsigned())

        # Main division loop.
        with m.If(in_progress):
            m.d.sync += iters_left.eq(iters_left - 1)

            shift_amt = (1 << (iters_left - 1).as_unsigned())
            with m.If(self.inp.payload.sign == Sign.SIGNED):
                m.d.comb += [
                    mag.divisor.eq(self.inp.payload.d.as_signed() * shift_amt),
                    mag.dividend.eq(remainder.as_signed())
                ]
            with m.Else():
                m.d.comb += [
                    mag.divisor.eq(self.inp.payload.d.as_unsigned() * shift_amt),  # noqa: E501
                    mag.dividend.eq(remainder)
                ]

            with m.If(mag.o):
                # If dividend/divisor are positive, subtract a positive
                # shifted divisor from dividend.
                with m.If((s_copy == Sign.UNSIGNED) | (~a_sign & ~n_sign)):
                    m.d.sync += quotient.eq(quotient + C(1) * shift_amt)
                    with m.If(s_copy == Sign.SIGNED):
                        # If high bit is set, and, a signed div,
                        # we want to sign-extend.
                        m.d.sync += remainder.eq(remainder -
                                                 (self.inp.payload.d.as_signed() * C(1) * shift_amt).as_signed())  # noqa: E501
                    with m.Else():
                        # Otherwise, zero-extend.
                        m.d.sync += remainder.eq(remainder -
                                                 (self.inp.payload.d.as_unsigned() * C(1) * shift_amt))  # noqa: E501
                # If dividend is negative, but divisor is positive, create a
                # negative shifted divisor and subtract from the dividend.
                with m.If((s_copy == Sign.SIGNED) & a_sign & ~n_sign): 
                    m.d.sync += quotient.eq(quotient + C(-1) * shift_amt)
                    m.d.sync += remainder.eq(remainder -
                                             (self.inp.payload.d.as_signed() * C(-1) * shift_amt).as_signed())  # noqa: E501
                # If dividend is positive, but divisor is negative, create a
                # positive shifted divisor and subtract from the dividend.
                with m.If((s_copy == Sign.SIGNED) & ~a_sign & n_sign):
                    m.d.sync += quotient.eq(quotient + C(-1) * shift_amt)
                    m.d.sync += remainder.eq(remainder -
                                             (self.inp.payload.d.as_signed() * C(-1) * shift_amt).as_signed())  # noqa: E501
                # If dividend/divisor are negative, subtract a negative
                # shifted divisor from dividend.
                with m.If((s_copy == Sign.SIGNED) & a_sign & n_sign):
                    m.d.sync += quotient.eq(quotient + C(1) * shift_amt)
                    m.d.sync += remainder.eq(remainder -
                                             (self.inp.payload.d.as_signed() * C(1) * shift_amt).as_signed())  # noqa: E501

            with m.If(iters_left - 1 == 0):
                m.d.sync += self.outp.valid.eq(1)

        return m


class _MagnitudeComparator(Elaboratable):
    """Compare the magnitude of two signed numbers."""

    def __init__(self, width=8):  # noqa: DOC
        self.width = width
        self.dividend = Signal(signed(self.width))
        self.divisor = Signal(signed(self.width))
        self.o = Signal()

    def elaborate(self, platform):
        m = Module()

        m.d.comb += self.o.eq(abs(self.divisor) <= abs(self.dividend))

        return m



import pytest
from smolarith.mul import PipelinedMul, Sign
from collections import deque
from itertools import product
import random


def mk_pipelined_testbench(m, abs_iter):
    def testbench():
        # Pipeline previous inputs... inputs to prev.append() are what
        # will go into the multiplier at the next active edge.
        # Outputs from prev.popleft()  are what went into the multiplier
        # "m.width - 1" active edges ago. This leads to a latency of
        # "m.width" clock cycles/ticks since the multiplier saw the inputs.
        prev = deque([(0, 0)]*m.width)

        # Not yielding to the simulator when values don't change
        # can result in significant speedups.
        a_prev = None
        b_prev = None
        s_prev = None

        yield m.inp.valid.eq(1)
        yield m.outp.ready.eq(1)

        for (a, b, s) in abs_iter:
            if s == Sign.UNSIGNED:
                if a != a_prev:
                    yield m.inp.payload.a.eq(a)
                if b != b_prev:
                    yield m.inp.payload.b.eq(b)
            else:
                if a != a_prev:
                    yield m.inp.payload.a.as_signed().eq(a)
                if b != b_prev:
                    yield m.inp.payload.b.as_signed().eq(b)

            if s != s_prev:
                yield m.inp.payload.sign.eq(s)
            yield
            (a_prev, b_prev, s_prev) = (a, b, s)

            (a_c, b_c) = prev.popleft()
            prev.append((a, b))

            if (yield m.outp.payload.sign) == Sign.UNSIGNED.value:
                assert a_c*b_c == (yield m.outp.payload.o)
            else:
                assert a_c*b_c == (yield m.outp.payload.o.as_signed())

            # print((a, b), (a_c, b_c), a_c*b_c, (yield m.o))
            # for i in range(8):
            #     print(f"{yield m.pin[i].a:08b}, {yield m.pin[i].b:08b}")
            # for i in range(8):
            #      print(f"{yield m.pout[i]:016b}")

        # Drain pipeline.
        for _ in range(m.width):
            yield
            (a_c, b_c) = prev.popleft()
            prev.append((a, b))

            if (yield m.outp.payload.sign) == Sign.UNSIGNED.value:
                assert a_c*b_c == (yield m.outp.payload.o)
            else:
                assert a_c*b_c == (yield m.outp.payload.o.as_signed())

    return testbench


@pytest.fixture
def all_values_tb(request, sim_mod):
    _, m = sim_mod
    mode = request.param

    if mode == Sign.UNSIGNED:
        a_range = range(0, 2**m.width)
        b_range = range(0, 2**m.width)
    elif mode == Sign.SIGNED:
        a_range = range(-2**(m.width-1), 2**(m.width-1))
        b_range = range(-2**(m.width-1), 2**(m.width-1))
    else:  # Sign.SIGNED_UNSIGNED
        a_range = range(-2**(m.width-1), 2**(m.width-1))
        b_range = range(0, 2**m.width)

    return mk_pipelined_testbench(m, product(a_range, b_range, (mode,)))


@pytest.fixture
def random_tb(sim_mod):
    _, m = sim_mod

    def random_muls():
        for i in range(256):
            a = random.randint(-2**31, (2**31)-1)
            b = random.randint(-2**31, (2**31)-1)
            s = random.choice([Sign.UNSIGNED, Sign.SIGNED,
                               Sign.SIGNED_UNSIGNED])

            if s == Sign.UNSIGNED:
                if a < 0:
                    a *= -1
                if b < 0:
                    b *= -1
            elif s == Sign.SIGNED_UNSIGNED and b < 0:
                b *= -1

            yield (a, b, s)

    return mk_pipelined_testbench(m, random_muls())


@pytest.fixture
def pipeline_tb(sim_mod):
    def testbench():
        _, m = sim_mod

        yield m.inp.payload.sign.eq(Sign.UNSIGNED)
        yield m.inp.payload.a.eq(1)
        yield m.inp.payload.b.eq(1)
        yield m.inp.valid.eq(1)
        yield m.outp.ready.eq(1)
        yield

        yield m.inp.payload.a.eq(2)
        yield m.inp.payload.b.eq(2)
        yield

        # Pipeline should continue working...
        yield m.inp.valid.eq(0)
        yield

        yield m.inp.valid.eq(1)
        yield m.inp.payload.a.eq(3)
        yield m.inp.payload.b.eq(3)
        yield

        yield m.inp.payload.a.eq(4)
        yield m.inp.payload.b.eq(4)
        yield

        yield m.inp.valid.eq(0)
        for _ in range(3):
            yield

        yield m.outp.ready.eq(0)
        yield

        yield m.outp.ready.eq(1)
        assert (yield m.outp.valid) == 1
        assert (yield m.outp.payload.sign) == Sign.UNSIGNED
        assert (yield m.outp.payload.o) == 1
        # Pipeline should stall...
        assert (yield m.inp.ready) == 0
        yield

        assert (yield m.outp.valid) == 1
        assert (yield m.outp.payload.sign) == Sign.UNSIGNED
        assert (yield m.outp.payload.o) == 1
        assert (yield m.inp.ready) == 1
        yield

        assert (yield m.outp.valid) == 1
        assert (yield m.outp.payload.sign) == Sign.UNSIGNED
        assert (yield m.outp.payload.o) == 4
        assert (yield m.inp.ready) == 1
        yield

        assert (yield m.outp.valid) == 0
        assert (yield m.inp.ready) == 1
        yield

        assert (yield m.outp.valid) == 1
        assert (yield m.outp.payload.sign) == Sign.UNSIGNED
        assert (yield m.outp.payload.o) == 9
        assert (yield m.inp.ready) == 1
        yield

        assert (yield m.outp.valid) == 1
        assert (yield m.outp.payload.sign) == Sign.UNSIGNED
        assert (yield m.outp.payload.o) == 16
        assert (yield m.inp.ready) == 1
        yield

        assert (yield m.outp.valid) == 0
        assert (yield m.inp.ready) == 1
        yield

    return testbench


@pytest.mark.module(PipelinedMul(8, debug=True))
@pytest.mark.clks((1.0 / 12e6,))
@pytest.mark.parametrize("all_values_tb", [Sign.UNSIGNED, Sign.SIGNED,
                                           Sign.SIGNED_UNSIGNED],
                         indirect=True)
def test_pipelined_mul(sim_mod, all_values_tb):
    sim, m = sim_mod
    sim.run(sync_processes=[all_values_tb])


@pytest.mark.module(PipelinedMul(32, debug=True))
@pytest.mark.clks((1.0 / 12e6,))
def test_random_32b(sim_mod, random_tb):
    sim, m = sim_mod
    random.seed(0)
    sim.run(sync_processes=[random_tb])


@pytest.mark.module(PipelinedMul(8, debug=True))
@pytest.mark.clks((1.0 / 12e6,))
def test_pipeline_stall(sim_mod, pipeline_tb):
    sim, m = sim_mod
    sim.run(sync_processes=[pipeline_tb])
